from typing import List

from wizwalker.memory.memory_object import DynamicMemoryObject, PropertyClass
from .enums import DelayOrder, SpellSourceType
from .spell_effect import DynamicSpellEffect


class SpellTemplate(PropertyClass):
    async def read_base_address(self) -> int:
        raise NotImplementedError()

    # async def behaviors(self) -> class BehaviorTemplate*:
    #     return await self.read_value_from_offset(72, "class BehaviorTemplate*")

    async def name(self) -> str:
        return await self.read_string_from_offset(96)

    async def write_name(self, name: str):
        await self.write_string_to_offset(96, name)

    async def description(self) -> str:
        return await self.read_string_from_offset(168)

    async def write_description(self, description: str):
        await self.write_string_to_offset(168, description)

    async def display_name(self) -> str:
        return await self.read_string_from_offset(136)

    async def write_display_name(self, display_name: str):
        await self.write_string_to_offset(136, display_name)

    async def spell_base(self) -> str:
        return await self.read_string_from_offset(208)

    async def write_spell_base(self, spell_base: str):
        await self.write_string_to_offset(208, spell_base)

    async def effects(self) -> List[DynamicSpellEffect]:
        effects = []
        for addr in await self.read_shared_vector(240):
            effects.append(DynamicSpellEffect(self.hook_handler, addr))

        return effects

    async def magic_school_name(self) -> str:
        return await self.read_string_from_offset(272)

    async def write_magic_school_name(self, magic_school_name: str):
        await self.write_string_to_offset(272, magic_school_name)

    async def type_name(self) -> str:
        return await self.read_string_from_offset(312)

    async def write_type_name(self, type_name: str):
        await self.write_string_to_offset(312, type_name)

    async def training_cost(self) -> int:
        return await self.read_value_from_offset(344, "int")

    async def write_training_cost(self, training_cost: int):
        await self.write_value_to_offset(344, training_cost, "int")

    async def accuracy(self) -> int:
        return await self.read_value_from_offset(348, "int")

    async def write_accuracy(self, accuracy: int):
        await self.write_value_to_offset(348, accuracy, "int")

    async def base_cost(self) -> int:
        return await self.read_value_from_offset(200, "int")

    async def write_base_cost(self, base_cost: int):
        await self.write_value_to_offset(200, base_cost, "int")

    async def credits_cost(self) -> int:
        return await self.read_value_from_offset(204, "int")

    async def write_credits_cost(self, credits_cost: int):
        await self.write_value_to_offset(204, credits_cost, "int")

    async def booster_pack_icon(self) -> str:
        return await self.read_string_from_offset(456)

    async def write_booster_pack_icon(self, booster_pack_icon: str):
        await self.write_string_to_offset(456, booster_pack_icon)

    async def valid_target_spells(self) -> int:
        return await self.read_value_from_offset(352, "unsigned int")

    async def write_valid_target_spells(self, valid_target_spells: int):
        await self.write_value_to_offset(352, valid_target_spells, "unsigned int")

    async def pvp(self) -> bool:
        return await self.read_value_from_offset(368, "bool")

    async def write_pvp(self, pvp: bool):
        await self.write_value_to_offset(368, pvp, "bool")

    async def pve(self) -> bool:
        return await self.read_value_from_offset(369, "bool")

    async def write_pve(self, _pv_e: bool):
        await self.write_value_to_offset(369, _pv_e, "bool")

    async def battlegrounds_only(self) -> bool:
        return await self.read_value_from_offset(370, "bool")

    async def write_battlegrounds_only(self, battlegrounds_only: bool):
        await self.write_value_to_offset(370, battlegrounds_only, "bool")

    async def treasure(self) -> bool:
        return await self.read_value_from_offset(371, "bool")

    async def write_treasure(self, treasure: bool):
        await self.write_value_to_offset(371, treasure, "bool")

    async def no_discard(self) -> bool:
        return await self.read_value_from_offset(372, "bool")

    async def write_no_discard(self, no_discard: bool):
        await self.write_value_to_offset(372, no_discard, "bool")

    async def leaves_play_when_cast(self) -> bool:
        return await self.read_value_from_offset(492, "bool")

    async def write_leaves_play_when_cast(self, leaves_play_when_cast: bool):
        await self.write_value_to_offset(492, leaves_play_when_cast, "bool")

    async def image_index(self) -> int:
        return await self.read_value_from_offset(376, "int")

    async def write_image_index(self, image_index: int):
        await self.write_value_to_offset(376, image_index, "int")

    async def image_name(self) -> str:
        return await self.read_string_from_offset(384)

    async def write_image_name(self, image_name: str):
        await self.write_string_to_offset(384, image_name)

    async def cloaked(self) -> bool:
        return await self.read_value_from_offset(449, "bool")

    async def write_cloaked(self, cloaked: bool):
        await self.write_value_to_offset(449, cloaked, "bool")

    async def caster_invisible(self) -> bool:
        return await self.read_value_from_offset(450, "bool")

    async def write_caster_invisible(self, caster_invisible: bool):
        await self.write_value_to_offset(450, caster_invisible, "bool")

    async def adjectives(self) -> str:
        return await self.read_string_from_offset(536)

    async def write_adjectives(self, adjectives: str):
        await self.write_string_to_offset(536, adjectives)

    async def spell_source_type(self) -> SpellSourceType:
        return await self.read_enum(488, SpellSourceType)

    async def write_spell_source_type(self, spell_source_type: SpellSourceType):
        await self.write_enum(488, spell_source_type)

    async def cloaked_name(self) -> str:
        return await self.read_string_from_offset(496)

    async def write_cloaked_name(self, cloaked_name: str):
        await self.write_string_to_offset(496, cloaked_name)

    # async def purchase_requirements(self) -> class RequirementList*:
    #     return await self.read_value_from_offset(568, "class RequirementList*")

    async def description_trainer(self) -> str:
        return await self.read_string_from_offset(576)

    async def write_description_trainer(self, description_trainer: str):
        await self.write_string_to_offset(576, description_trainer)

    async def description_combat_hud(self) -> str:
        return await self.read_string_from_offset(608)

    async def write_description_combat_hud(self, description_combat_hud: str):
        await self.write_string_to_offset(608, description_combat_hud)

    async def display_index(self) -> int:
        return await self.read_value_from_offset(640, "int")

    async def write_display_index(self, display_index: int):
        await self.write_value_to_offset(640, display_index, "int")

    async def hidden_from_effects_window(self) -> bool:
        return await self.read_value_from_offset(644, "bool")

    async def write_hidden_from_effects_window(self, hidden_from_effects_window: bool):
        await self.write_value_to_offset(644, hidden_from_effects_window, "bool")

    async def ignore_charms(self) -> bool:
        return await self.read_value_from_offset(645, "bool")

    async def write_ignore_charms(self, ignore_charms: bool):
        await self.write_value_to_offset(645, ignore_charms, "bool")

    async def always_fizzle(self) -> bool:
        return await self.read_value_from_offset(646, "bool")

    async def write_always_fizzle(self, always_fizzle: bool):
        await self.write_value_to_offset(646, always_fizzle, "bool")

    async def spell_category(self) -> str:
        return await self.read_string_from_offset(648)

    async def write_spell_category(self, spell_category: str):
        await self.write_string_to_offset(648, spell_category)

    async def show_polymorphed_name(self) -> bool:
        return await self.read_value_from_offset(680, "bool")

    async def write_show_polymorphed_name(self, show_polymorphed_name: bool):
        await self.write_value_to_offset(680, show_polymorphed_name, "bool")

    async def skip_truncation(self) -> bool:
        return await self.read_value_from_offset(681, "bool")

    async def write_skip_truncation(self, skip_truncation: bool):
        await self.write_value_to_offset(681, skip_truncation, "bool")

    async def max_copies(self) -> int:
        return await self.read_value_from_offset(684, "unsigned int")

    async def write_max_copies(self, max_copies: int):
        await self.write_value_to_offset(684, max_copies, "unsigned int")

    async def level_restriction(self) -> int:
        return await self.read_value_from_offset(688, "int")

    async def write_level_restriction(self, level_restriction: int):
        await self.write_value_to_offset(688, level_restriction, "int")

    async def delay_enchantment(self) -> bool:
        return await self.read_value_from_offset(692, "bool")

    async def write_delay_enchantment(self, delay_enchantment: bool):
        await self.write_value_to_offset(692, delay_enchantment, "bool")

    async def delay_enchantment_order(self) -> DelayOrder:
        return await self.read_enum(696, DelayOrder)

    async def write_delay_enchantment_order(self, delay_enchantment_order: DelayOrder):
        await self.write_enum(696, delay_enchantment_order)

    async def previous_spell_name(self) -> str:
        return await self.read_string_from_offset(704)

    async def write_previous_spell_name(self, previous_spell_name: str):
        await self.write_string_to_offset(704, previous_spell_name)

    async def card_front(self) -> str:
        return await self.read_string_from_offset(416)

    async def write_card_front(self, card_front: str):
        await self.write_string_to_offset(416, card_front)

    async def use_gloss(self) -> bool:
        return await self.read_value_from_offset(448, "bool")

    async def write_use_gloss(self, use_gloss: bool):
        await self.write_value_to_offset(448, use_gloss, "bool")


class DynamicSpellTemplate(DynamicMemoryObject, SpellTemplate):
    pass



from typing import List

from wizwalker.memory.memory_object import PropertyClass, DynamicMemoryObject
from .spell_effect import DynamicSpellEffect


class CombatResolver(PropertyClass):
    async def read_base_address(self) -> int:
        raise NotImplementedError()

    async def bool_global_effect(self) -> bool:
        return await self.read_value_from_offset(112, "bool")

    async def write_bool_global_effect(self, bool_global_effect: bool):
        await self.write_value_to_offset(112, bool_global_effect, "bool")

    async def global_effect(self) -> DynamicSpellEffect:
        addr = await self.read_value_from_offset(120, "long long")
        return DynamicSpellEffect(self.hook_handler, addr)

    async def battlefield_effects(self) -> List[DynamicSpellEffect]:
        effects = []
        for addr in await self.read_shared_vector(136):
            effects.append(DynamicSpellEffect(self.hook_handler, addr))

        return effects


class DynamicCombatResolver(DynamicMemoryObject, CombatResolver):
    pass



from .actor_body import ActorBody, CurrentActorBody
from .client_zone import ClientZone, DynamicClientZone
from .client_object import (
    ClientObject,
    CurrentClientObject,
    DynamicClientObject,
)
from .combat_participant import CombatParticipant, DynamicCombatParticipant
from .duel import CurrentDuel, Duel
from .enums import *
from .game_stats import CurrentGameStats
from .quest_position import CurrentQuestPosition
from .spell_effect import SpellEffects, DynamicSpellEffect
from .spell_template import SpellTemplate, DynamicSpellTemplate
from .spell import DynamicHand, DynamicSpell, Hand, Spell
from .window import CurrentRootWindow, DynamicWindow, Window
from .render_context import RenderContext, CurrentRenderContext
from .combat_resolver import CombatResolver, DynamicCombatResolver
from .play_deck import PlayDeck, PlaySpellData, DynamicPlayDeck, DynamicPlaySpellData
from .game_object_template import WizGameObjectTemplate, DynamicWizGameObjectTemplate
from .behavior_template import BehaviorTemplate, DynamicBehaviorTemplate
from .behavior_instance import BehaviorInstance, DynamicBehaviorInstance



from wizwalker.memory.memory_object import DynamicMemoryObject, PropertyClass


class GameStats(PropertyClass):
    async def read_base_address(self) -> int:
        raise NotImplementedError()

    async def max_hitpoints(self) -> int:
        """
        Client's max hitpoints; base + bonus
        """
        base = await self.base_hitpoints()
        bonus = await self.bonus_hitpoints()
        return base + bonus

    async def max_mana(self) -> int:
        """
        Clients's max mana; base + bonus
        """
        base = await self.base_mana()
        bonus = await self.bonus_mana()
        return base + bonus

    async def base_hitpoints(self) -> int:
        return await self.read_value_from_offset(80, "int")

    async def write_base_hitpoints(self, base_hitpoints: int):
        await self.write_value_to_offset(80, base_hitpoints, "int")

    async def base_mana(self) -> int:
        return await self.read_value_from_offset(84, "int")

    async def write_base_mana(self, base_mana: int):
        await self.write_value_to_offset(84, base_mana, "int")

    async def base_gold_pouch(self) -> int:
        return await self.read_value_from_offset(88, "int")

    async def write_base_gold_pouch(self, base_gold_pouch: int):
        await self.write_value_to_offset(88, base_gold_pouch, "int")

    async def base_event_currency1_pouch(self) -> int:
        return await self.read_value_from_offset(92, "int")

    async def write_base_event_currency1_pouch(self, base_event_currency1_pouch: int):
        await self.write_value_to_offset(92, base_event_currency1_pouch, "int")

    async def base_event_currency2_pouch(self) -> int:
        return await self.read_value_from_offset(96, "int")

    async def write_base_event_currency2_pouch(self, base_event_currency2_pouch: int):
        await self.write_value_to_offset(96, base_event_currency2_pouch, "int")

    async def energy_max(self) -> int:
        return await self.read_value_from_offset(100, "int")

    async def write_energy_max(self, energy_max: int):
        await self.write_value_to_offset(100, energy_max, "int")

    async def current_hitpoints(self) -> int:
        return await self.read_value_from_offset(104, "int")

    async def write_current_hitpoints(self, current_hitpoints: int):
        await self.write_value_to_offset(104, current_hitpoints, "int")

    async def current_gold(self) -> int:
        return await self.read_value_from_offset(108, "int")

    async def write_current_gold(self, current_gold: int):
        await self.write_value_to_offset(108, current_gold, "int")

    async def current_event_currency1(self) -> int:
        return await self.read_value_from_offset(112, "int")

    async def write_current_event_currency1(self, current_event_currency1: int):
        await self.write_value_to_offset(112, current_event_currency1, "int")

    async def current_event_currency2(self) -> int:
        return await self.read_value_from_offset(116, "int")

    async def write_current_event_currency2(self, current_event_currency2: int):
        await self.write_value_to_offset(116, current_event_currency2, "int")

    async def current_mana(self) -> int:
        return await self.read_value_from_offset(120, "int")

    async def write_current_mana(self, current_mana: int):
        await self.write_value_to_offset(120, current_mana, "int")

    async def current_arena_points(self) -> int:
        return await self.read_value_from_offset(124, "int")

    async def write_current_arena_points(self, current_arena_points: int):
        await self.write_value_to_offset(124, current_arena_points, "int")

    async def spell_charge_base(self) -> int:
        return await self.read_value_from_offset(128, "int")

    async def write_spell_charge_base(self, spell_charge_base: int):
        await self.write_value_to_offset(128, spell_charge_base, "int")

    async def potion_max(self) -> float:
        return await self.read_value_from_offset(152, "float")

    async def write_potion_max(self, potion_max: float):
        await self.write_value_to_offset(152, potion_max, "float")

    async def potion_charge(self) -> float:
        return await self.read_value_from_offset(156, "float")

    async def write_potion_charge(self, potion_charge: float):
        await self.write_value_to_offset(156, potion_charge, "float")

    # async def arena_ladder(self) -> class SharedPointer<class Ladder>:
    #     return await self.read_value_from_offset(160, "class SharedPointer<class Ladder>")

    # async def derby_ladder(self) -> class SharedPointer<class Ladder>:
    #     return await self.read_value_from_offset(176, "class SharedPointer<class Ladder>")

    # async def bracket_lader(self) -> class SharedPointer<class Ladder>:
    #     return await self.read_value_from_offset(192, "class SharedPointer<class Ladder>")

    async def bonus_hitpoints(self) -> int:
        return await self.read_value_from_offset(208, "int")

    async def write_bonus_hitpoints(self, bonus_hitpoints: int):
        await self.write_value_to_offset(208, bonus_hitpoints, "int")

    async def bonus_mana(self) -> int:
        return await self.read_value_from_offset(212, "int")

    async def write_bonus_mana(self, bonus_mana: int):
        await self.write_value_to_offset(212, bonus_mana, "int")

    async def bonus_energy(self) -> int:
        return await self.read_value_from_offset(228, "int")

    async def write_bonus_energy(self, bonus_energy: int):
        await self.write_value_to_offset(228, bonus_energy, "int")

    async def critical_hit_percent_all(self) -> float:
        return await self.read_value_from_offset(232, "float")

    async def write_critical_hit_percent_all(self, critical_hit_percent_all: float):
        await self.write_value_to_offset(232, critical_hit_percent_all, "float")

    async def block_percent_all(self) -> float:
        return await self.read_value_from_offset(236, "float")

    async def write_block_percent_all(self, block_percent_all: float):
        await self.write_value_to_offset(236, block_percent_all, "float")

    async def critical_hit_rating_all(self) -> float:
        return await self.read_value_from_offset(240, "float")

    async def write_critical_hit_rating_all(self, critical_hit_rating_all: float):
        await self.write_value_to_offset(240, critical_hit_rating_all, "float")

    async def block_rating_all(self) -> float:
        return await self.read_value_from_offset(244, "float")

    async def write_block_rating_all(self, block_rating_all: float):
        await self.write_value_to_offset(244, block_rating_all, "float")

    async def reference_level(self) -> int:
        return await self.read_value_from_offset(308, "int")

    async def write_reference_level(self, reference_level: int):
        await self.write_value_to_offset(308, reference_level, "int")

    async def highest_character_level_on_account(self) -> int:
        return await self.read_value_from_offset(312, "int")

    async def write_highest_character_level_on_account(
        self, highest_character_level_on_account: int
    ):
        await self.write_value_to_offset(312, highest_character_level_on_account, "int")

    async def pet_act_chance(self) -> int:
        return await self.read_value_from_offset(316, "int")

    async def write_pet_act_chance(self, pet_act_chance: int):
        await self.write_value_to_offset(316, pet_act_chance, "int")

    async def dmg_bonus_percent(self) -> float:
        return await self.read_value_from_offset(320, "float")

    async def write_dmg_bonus_percent(self, dmg_bonus_percent: float):
        await self.write_value_to_offset(320, dmg_bonus_percent, "float")

    async def dmg_bonus_flat(self) -> float:
        return await self.read_value_from_offset(344, "float")

    async def write_dmg_bonus_flat(self, dmg_bonus_flat: float):
        await self.write_value_to_offset(344, dmg_bonus_flat, "float")

    async def acc_bonus_percent(self) -> float:
        return await self.read_value_from_offset(368, "float")

    async def write_acc_bonus_percent(self, acc_bonus_percent: float):
        await self.write_value_to_offset(368, acc_bonus_percent, "float")

    async def ap_bonus_percent(self) -> float:
        return await self.read_value_from_offset(392, "float")

    async def write_ap_bonus_percent(self, ap_bonus_percent: float):
        await self.write_value_to_offset(392, ap_bonus_percent, "float")

    async def dmg_reduce_percent(self) -> float:
        return await self.read_value_from_offset(416, "float")

    async def write_dmg_reduce_percent(self, dmg_reduce_percent: float):
        await self.write_value_to_offset(416, dmg_reduce_percent, "float")

    async def dmg_reduce_flat(self) -> float:
        return await self.read_value_from_offset(440, "float")

    async def write_dmg_reduce_flat(self, dmg_reduce_flat: float):
        await self.write_value_to_offset(440, dmg_reduce_flat, "float")

    async def acc_reduce_percent(self) -> float:
        return await self.read_value_from_offset(464, "float")

    async def write_acc_reduce_percent(self, acc_reduce_percent: float):
        await self.write_value_to_offset(464, acc_reduce_percent, "float")

    async def heal_bonus_percent(self) -> float:
        return await self.read_value_from_offset(488, "float")

    async def write_heal_bonus_percent(self, heal_bonus_percent: float):
        await self.write_value_to_offset(488, heal_bonus_percent, "float")

    async def heal_inc_bonus_percent(self) -> float:
        return await self.read_value_from_offset(512, "float")

    async def write_heal_inc_bonus_percent(self, heal_inc_bonus_percent: float):
        await self.write_value_to_offset(512, heal_inc_bonus_percent, "float")

    async def spell_charge_bonus(self) -> int:
        return await self.read_value_from_offset(560, "int")

    async def write_spell_charge_bonus(self, spell_charge_bonus: int):
        await self.write_value_to_offset(560, spell_charge_bonus, "int")

    async def dmg_bonus_percent_all(self) -> float:
        return await self.read_value_from_offset(680, "float")

    async def write_dmg_bonus_percent_all(self, dmg_bonus_percent_all: float):
        await self.write_value_to_offset(680, dmg_bonus_percent_all, "float")

    async def dmg_bonus_flat_all(self) -> float:
        return await self.read_value_from_offset(684, "float")

    async def write_dmg_bonus_flat_all(self, dmg_bonus_flat_all: float):
        await self.write_value_to_offset(684, dmg_bonus_flat_all, "float")

    async def acc_bonus_percent_all(self) -> float:
        return await self.read_value_from_offset(688, "float")

    async def write_acc_bonus_percent_all(self, acc_bonus_percent_all: float):
        await self.write_value_to_offset(688, acc_bonus_percent_all, "float")

    async def ap_bonus_percent_all(self) -> float:
        return await self.read_value_from_offset(692, "float")

    async def write_ap_bonus_percent_all(self, ap_bonus_percent_all: float):
        await self.write_value_to_offset(692, ap_bonus_percent_all, "float")

    async def dmg_reduce_percent_all(self) -> float:
        return await self.read_value_from_offset(696, "float")

    async def write_dmg_reduce_percent_all(self, dmg_reduce_percent_all: float):
        await self.write_value_to_offset(696, dmg_reduce_percent_all, "float")

    async def dmg_reduce_flat_all(self) -> float:
        return await self.read_value_from_offset(700, "float")

    async def write_dmg_reduce_flat_all(self, dmg_reduce_flat_all: float):
        await self.write_value_to_offset(700, dmg_reduce_flat_all, "float")

    async def acc_reduce_percent_all(self) -> float:
        return await self.read_value_from_offset(704, "float")

    async def write_acc_reduce_percent_all(self, acc_reduce_percent_all: float):
        await self.write_value_to_offset(704, acc_reduce_percent_all, "float")

    async def heal_bonus_percent_all(self) -> float:
        return await self.read_value_from_offset(708, "float")

    async def write_heal_bonus_percent_all(self, heal_bonus_percent_all: float):
        await self.write_value_to_offset(708, heal_bonus_percent_all, "float")

    async def heal_inc_bonus_percent_all(self) -> float:
        return await self.read_value_from_offset(712, "float")

    async def write_heal_inc_bonus_percent_all(self, heal_inc_bonus_percent_all: float):
        await self.write_value_to_offset(712, heal_inc_bonus_percent_all, "float")

    async def spell_charge_bonus_all(self) -> int:
        return await self.read_value_from_offset(720, "int")

    async def write_spell_charge_bonus_all(self, spell_charge_bonus_all: int):
        await self.write_value_to_offset(720, spell_charge_bonus_all, "int")

    async def power_pip_base(self) -> float:
        return await self.read_value_from_offset(724, "float")

    async def write_power_pip_base(self, power_pip_base: float):
        await self.write_value_to_offset(724, power_pip_base, "float")

    async def power_pip_bonus_percent_all(self) -> float:
        return await self.read_value_from_offset(760, "float")

    async def write_power_pip_bonus_percent_all(
        self, power_pip_bonus_percent_all: float
    ):
        await self.write_value_to_offset(760, power_pip_bonus_percent_all, "float")

    async def xp_percent_increase(self) -> float:
        return await self.read_value_from_offset(768, "float")

    async def write_xp_percent_increase(self, xp_percent_increase: float):
        await self.write_value_to_offset(768, xp_percent_increase, "float")

    async def critical_hit_percent_by_school(self) -> float:
        return await self.read_value_from_offset(584, "float")

    async def write_critical_hit_percent_by_school(
        self, critical_hit_percent_by_school: float
    ):
        await self.write_value_to_offset(584, critical_hit_percent_by_school, "float")

    async def block_percent_by_school(self) -> float:
        return await self.read_value_from_offset(608, "float")

    async def write_block_percent_by_school(self, block_percent_by_school: float):
        await self.write_value_to_offset(608, block_percent_by_school, "float")

    async def critical_hit_rating_by_school(self) -> float:
        return await self.read_value_from_offset(632, "float")

    async def write_critical_hit_rating_by_school(
        self, critical_hit_rating_by_school: float
    ):
        await self.write_value_to_offset(632, critical_hit_rating_by_school, "float")

    async def block_rating_by_school(self) -> float:
        return await self.read_value_from_offset(656, "float")

    async def write_block_rating_by_school(self, block_rating_by_school: float):
        await self.write_value_to_offset(656, block_rating_by_school, "float")

    async def balance_mastery(self) -> int:
        return await self.read_value_from_offset(792, "int")

    async def write_balance_mastery(self, balance_mastery: int):
        await self.write_value_to_offset(792, balance_mastery, "int")

    async def death_mastery(self) -> int:
        return await self.read_value_from_offset(796, "int")

    async def write_death_mastery(self, death_mastery: int):
        await self.write_value_to_offset(796, death_mastery, "int")

    async def fire_mastery(self) -> int:
        return await self.read_value_from_offset(800, "int")

    async def write_fire_mastery(self, fire_mastery: int):
        await self.write_value_to_offset(800, fire_mastery, "int")

    async def ice_mastery(self) -> int:
        return await self.read_value_from_offset(804, "int")

    async def write_ice_mastery(self, ice_mastery: int):
        await self.write_value_to_offset(804, ice_mastery, "int")

    async def life_mastery(self) -> int:
        return await self.read_value_from_offset(808, "int")

    async def write_life_mastery(self, life_mastery: int):
        await self.write_value_to_offset(808, life_mastery, "int")

    async def myth_mastery(self) -> int:
        return await self.read_value_from_offset(812, "int")

    async def write_myth_mastery(self, myth_mastery: int):
        await self.write_value_to_offset(812, myth_mastery, "int")

    async def storm_mastery(self) -> int:
        return await self.read_value_from_offset(816, "int")

    async def write_storm_mastery(self, storm_mastery: int):
        await self.write_value_to_offset(816, storm_mastery, "int")

    async def maximum_number_of_islands(self) -> int:
        return await self.read_value_from_offset(820, "int")

    async def write_maximum_number_of_islands(self, maximum_number_of_islands: int):
        await self.write_value_to_offset(820, maximum_number_of_islands, "int")

    async def gardening_level(self) -> int:
        return await self.read_value_from_offset(824, "unsigned char")

    async def write_gardening_level(self, gardening_level: int):
        await self.write_value_to_offset(824, gardening_level, "unsigned char")

    async def gardening_xp(self) -> int:
        return await self.read_value_from_offset(828, "int")

    async def write_gardening_xp(self, gardening_xp: int):
        await self.write_value_to_offset(828, gardening_xp, "int")

    async def invisible_to_friends(self) -> bool:
        return await self.read_value_from_offset(832, "bool")

    async def write_invisible_to_friends(self, invisible_to_friends: bool):
        await self.write_value_to_offset(832, invisible_to_friends, "bool")

    async def show_item_lock(self) -> bool:
        return await self.read_value_from_offset(833, "bool")

    async def write_show_item_lock(self, show_item_lock: bool):
        await self.write_value_to_offset(833, show_item_lock, "bool")

    async def quest_finder_enabled(self) -> bool:
        return await self.read_value_from_offset(834, "bool")

    async def write_quest_finder_enabled(self, quest_finder_enabled: bool):
        await self.write_value_to_offset(834, quest_finder_enabled, "bool")

    async def buddy_list_limit(self) -> int:
        return await self.read_value_from_offset(836, "int")

    async def write_buddy_list_limit(self, buddy_list_limit: int):
        await self.write_value_to_offset(836, buddy_list_limit, "int")

    async def dont_allow_friend_finder_codes(self) -> bool:
        return await self.read_value_from_offset(844, "bool")

    async def write_dont_allow_friend_finder_codes(
        self, dont_allow_friend_finder_codes: bool
    ):
        await self.write_value_to_offset(844, dont_allow_friend_finder_codes, "bool")

    async def stun_resistance_percent(self) -> float:
        return await self.read_value_from_offset(840, "float")

    async def write_stun_resistance_percent(self, stun_resistance_percent: float):
        await self.write_value_to_offset(840, stun_resistance_percent, "float")

    async def shadow_magic_unlocked(self) -> bool:
        return await self.read_value_from_offset(852, "bool")

    async def write_shadow_magic_unlocked(self, shadow_magic_unlocked: bool):
        await self.write_value_to_offset(852, shadow_magic_unlocked, "bool")

    async def shadow_pip_max(self) -> int:
        return await self.read_value_from_offset(848, "int")

    async def write_shadow_pip_max(self, shadow_pip_max: int):
        await self.write_value_to_offset(848, shadow_pip_max, "int")

    async def fishing_level(self) -> int:
        return await self.read_value_from_offset(853, "unsigned char")

    async def write_fishing_level(self, fishing_level: int):
        await self.write_value_to_offset(853, fishing_level, "unsigned char")

    async def fishing_xp(self) -> int:
        return await self.read_value_from_offset(856, "int")

    async def write_fishing_xp(self, fishing_xp: int):
        await self.write_value_to_offset(856, fishing_xp, "int")

    async def fishing_luck_bonus_percent(self) -> float:
        return await self.read_value_from_offset(536, "float")

    async def write_fishing_luck_bonus_percent(self, fishing_luck_bonus_percent: float):
        await self.write_value_to_offset(536, fishing_luck_bonus_percent, "float")

    async def fishing_luck_bonus_percent_all(self) -> float:
        return await self.read_value_from_offset(716, "float")

    async def write_fishing_luck_bonus_percent_all(
        self, fishing_luck_bonus_percent_all: float
    ):
        await self.write_value_to_offset(716, fishing_luck_bonus_percent_all, "float")

    async def subscriber_benefit_flags(self) -> int:
        return await self.read_value_from_offset(860, "unsigned int")

    async def write_subscriber_benefit_flags(self, subscriber_benefit_flags: int):
        await self.write_value_to_offset(860, subscriber_benefit_flags, "unsigned int")

    async def elixir_benefit_flags(self) -> int:
        return await self.read_value_from_offset(864, "unsigned int")

    async def write_elixir_benefit_flags(self, elixir_benefit_flags: int):
        await self.write_value_to_offset(864, elixir_benefit_flags, "unsigned int")

    async def shadow_pip_bonus_percent(self) -> float:
        return await self.read_value_from_offset(764, "float")

    async def write_shadow_pip_bonus_percent(self, shadow_pip_bonus_percent: float):
        await self.write_value_to_offset(764, shadow_pip_bonus_percent, "float")

    async def wisp_bonus_percent(self) -> float:
        return await self.read_value_from_offset(784, "float")

    async def write_wisp_bonus_percent(self, wisp_bonus_percent: float):
        await self.write_value_to_offset(784, wisp_bonus_percent, "float")

    async def pip_conversion_rating_all(self) -> float:
        return await self.read_value_from_offset(272, "float")

    async def write_pip_conversion_rating_all(self, pip_conversion_rating_all: float):
        await self.write_value_to_offset(272, pip_conversion_rating_all, "float")

    async def pip_conversion_rating_per_school(self) -> float:
        return await self.read_value_from_offset(248, "float")

    async def write_pip_conversion_rating_per_school(
        self, pip_conversion_rating_per_school: float
    ):
        await self.write_value_to_offset(248, pip_conversion_rating_per_school, "float")

    async def pip_conversion_percent_all(self) -> float:
        return await self.read_value_from_offset(304, "float")

    async def write_pip_conversion_percent_all(self, pip_conversion_percent_all: float):
        await self.write_value_to_offset(304, pip_conversion_percent_all, "float")

    async def pip_conversion_percent_per_school(self) -> float:
        return await self.read_value_from_offset(280, "float")

    async def write_pip_conversion_percent_per_school(
        self, pip_conversion_percent_per_school: float
    ):
        await self.write_value_to_offset(
            280, pip_conversion_percent_per_school, "float"
        )

    async def monster_magic_level(self) -> int:
        return await self.read_value_from_offset(868, "unsigned char")

    async def write_monster_magic_level(self, monster_magic_level: int):
        await self.write_value_to_offset(868, monster_magic_level, "unsigned char")

    async def monster_magic_xp(self) -> int:
        return await self.read_value_from_offset(872, "int")

    async def write_monster_magic_xp(self, monster_magic_xp: int):
        await self.write_value_to_offset(872, monster_magic_xp, "int")

    async def player_chat_channel_is_public(self) -> bool:
        return await self.read_value_from_offset(876, "bool")

    async def write_player_chat_channel_is_public(
        self, player_chat_channel_is_public: bool
    ):
        await self.write_value_to_offset(876, player_chat_channel_is_public, "bool")

    async def extra_inventory_space(self) -> int:
        return await self.read_value_from_offset(880, "int")

    async def write_extra_inventory_space(self, extra_inventory_space: int):
        await self.write_value_to_offset(880, extra_inventory_space, "int")

    async def remember_last_realm(self) -> bool:
        return await self.read_value_from_offset(884, "bool")

    async def write_remember_last_realm(self, remember_last_realm: bool):
        await self.write_value_to_offset(884, remember_last_realm, "bool")

    async def new_spellbook_layout_warning(self) -> bool:
        return await self.read_value_from_offset(885, "bool")

    async def write_new_spellbook_layout_warning(
        self, new_spellbook_layout_warning: bool
    ):
        await self.write_value_to_offset(885, new_spellbook_layout_warning, "bool")

    async def pip_conversion_base_all_schools(self) -> int:
        return await self.read_value_from_offset(728, "int")

    async def write_pip_conversion_base_all_schools(
        self, pip_conversion_base_all_schools: int
    ):
        await self.write_value_to_offset(728, pip_conversion_base_all_schools, "int")

    async def pip_conversion_base_per_school(self) -> int:
        return await self.read_value_from_offset(736, "int")

    async def write_pip_conversion_base_per_school(
        self, pip_conversion_base_per_school: int
    ):
        await self.write_value_to_offset(736, pip_conversion_base_per_school, "int")

    async def purchased_custom_emotes1(self) -> int:
        return await self.read_value_from_offset(888, "unsigned int")

    async def write_purchased_custom_emotes1(self, purchased_custom_emotes1: int):
        await self.write_value_to_offset(888, purchased_custom_emotes1, "unsigned int")

    async def purchased_custom_teleport_effects1(self) -> int:
        return await self.read_value_from_offset(892, "unsigned int")

    async def write_purchased_custom_teleport_effects1(
        self, purchased_custom_teleport_effects1: int
    ):
        await self.write_value_to_offset(
            892, purchased_custom_teleport_effects1, "unsigned int"
        )

    async def equipped_teleport_effect(self) -> int:
        return await self.read_value_from_offset(896, "unsigned int")

    async def write_equipped_teleport_effect(self, equipped_teleport_effect: int):
        await self.write_value_to_offset(896, equipped_teleport_effect, "unsigned int")

    async def highest_world1_id(self) -> int:
        return await self.read_value_from_offset(900, "unsigned int")

    async def write_highest_world1_id(self, highest_world1_id: int):
        await self.write_value_to_offset(900, highest_world1_id, "unsigned int")

    async def highest_world2_id(self) -> int:
        return await self.read_value_from_offset(904, "unsigned int")

    async def write_highest_world2_id(self, highest_world2_i_d: int):
        await self.write_value_to_offset(904, highest_world2_i_d, "unsigned int")

    async def active_class_projects_list(self) -> int:
        return await self.read_value_from_offset(912, "unsigned int")

    async def write_active_class_projects_list(self, active_class_projects_list: int):
        await self.write_value_to_offset(
            912, active_class_projects_list, "unsigned int"
        )

    async def disabled_item_slot_ids(self) -> int:
        return await self.read_value_from_offset(928, "unsigned int")

    async def write_disabled_item_slot_ids(self, disabled_item_slot_ids: int):
        await self.write_value_to_offset(928, disabled_item_slot_ids, "unsigned int")

    async def adventure_power_cooldown_time(self) -> int:
        return await self.read_value_from_offset(944, "unsigned int")

    async def write_adventure_power_cooldown_time(
        self, adventure_power_cooldown_time: int
    ):
        await self.write_value_to_offset(
            944, adventure_power_cooldown_time, "unsigned int"
        )

    async def purchased_custom_emotes2(self) -> int:
        return await self.read_value_from_offset(948, "unsigned int")

    async def write_purchased_custom_emotes2(self, purchased_custom_emotes2: int):
        await self.write_value_to_offset(948, purchased_custom_emotes2, "unsigned int")

    async def purchased_custom_teleport_effects2(self) -> int:
        return await self.read_value_from_offset(952, "unsigned int")

    async def write_purchased_custom_teleport_effects2(
        self, purchased_custom_teleport_effects2: int
    ):
        await self.write_value_to_offset(
            952, purchased_custom_teleport_effects2, "unsigned int"
        )

    async def purchased_custom_emotes3(self) -> int:
        return await self.read_value_from_offset(956, "unsigned int")

    async def write_purchased_custom_emotes3(self, purchased_custom_emotes3: int):
        await self.write_value_to_offset(956, purchased_custom_emotes3, "unsigned int")

    async def purchased_custom_teleport_effects3(self) -> int:
        return await self.read_value_from_offset(960, "unsigned int")

    async def write_purchased_custom_teleport_effects3(
        self, purchased_custom_teleport_effects3: int
    ):
        await self.write_value_to_offset(
            960, purchased_custom_teleport_effects3, "unsigned int"
        )

    async def shadow_pip_rating(self) -> float:
        return await self.read_value_from_offset(964, "float")

    async def write_shadow_pip_rating(self, shadow_pip_rating: float):
        await self.write_value_to_offset(964, shadow_pip_rating, "float")

    async def bonus_shadow_pip_rating(self) -> float:
        return await self.read_value_from_offset(968, "float")

    async def write_bonus_shadow_pip_rating(self, bonus_shadow_pip_rating: float):
        await self.write_value_to_offset(968, bonus_shadow_pip_rating, "float")

    async def shadow_pip_rate_accumulated(self) -> float:
        return await self.read_value_from_offset(972, "float")

    async def write_shadow_pip_rate_accumulated(
        self, shadow_pip_rate_accumulated: float
    ):
        await self.write_value_to_offset(972, shadow_pip_rate_accumulated, "float")

    async def shadow_pip_rate_threshold(self) -> float:
        return await self.read_value_from_offset(976, "float")

    async def write_shadow_pip_rate_threshold(self, shadow_pip_rate_threshold: float):
        await self.write_value_to_offset(976, shadow_pip_rate_threshold, "float")

    async def shadow_pip_rate_percentage(self) -> int:
        return await self.read_value_from_offset(980, "int")

    async def write_shadow_pip_rate_percentage(self, shadow_pip_rate_percentage: int):
        await self.write_value_to_offset(980, shadow_pip_rate_percentage, "int")

    async def friendly_player(self) -> bool:
        return await self.read_value_from_offset(984, "bool")

    async def write_friendly_player(self, friendly_player: bool):
        await self.write_value_to_offset(984, friendly_player, "bool")


class CurrentGameStats(GameStats):
    async def read_base_address(self) -> int:
        return await self.hook_handler.read_current_player_stat_base()


class DynamicGameStats(DynamicMemoryObject, GameStats):
    pass



from enum import Enum, IntFlag


class DuelPhase(Enum):
    starting = 0
    pre_planning = 1
    planning = 2
    pre_execution = 3
    execution = 4
    resolution = 5
    victory = 6
    ended = 7
    max = 10


class SigilInitiativeSwitchMode(Enum):
    none = 0
    reroll = 1
    switch = 2


class DuelExecutionOrder(Enum):
    sequential = 0
    alternating = 1


class PipAquiredByEnum(Enum):
    unknown = 0
    normal = 1
    power = 2
    normal_to_power_conversion = 4
    impede_pips = 5


class DelayOrder(Enum):
    any_order = 0
    first = 1
    second = 2


class WindowStyle(IntFlag):
    has_back = 1
    scale_children = 2
    can_move = 4
    can_scroll = 16
    focus_locked = 64
    can_focus = 128
    can_dock = 32
    do_not_capture_mouse = 256
    is_transparent = 256
    effect_fadeid = 512
    effect_highlight = 1024
    has_no_border = 2048
    ignore_parent_scale = 4096
    use_alpha_bounds = 8192
    auto_grow = 16384
    auto_shrink = 32768
    auto_resize = 49152


class WindowFlags(IntFlag):
    visible = 1
    noclip = 2
    dock_outside = 131072
    dock_left = 128
    dock_top = 512
    dock_right = 256
    dock_bottom = 1024
    parent_size = 786432
    parent_width = 262144
    parent_height = 524288
    hcenter = 32768
    vcenter = 65536
    disabled = -2147483648


class SpellSourceType(Enum):
    caster = 0
    pet = 1
    shadow_creature = 2
    weapon = 3
    equipment = 4


class SpellEffects(Enum):
    invalid_spell_effect = 0
    damage = 1
    damage_no_crit = 2
    heal = 3
    heal_percent = 4
    set_heal_percent = 111
    steal_health = 5
    reduce_over_time = 6
    detonate_over_time = 7
    push_charm = 8
    steal_charm = 9
    push_ward = 10
    steal_ward = 11
    push_over_time = 12
    steal_over_time = 13
    remove_charm = 14
    remove_ward = 15
    remove_over_time = 16
    remove_aura = 17
    swap_all = 18
    swap_charm = 19
    swap_ward = 20
    swap_over_time = 21
    modify_incoming_damage = 22
    modify_incoming_damage_flat = 117
    maximum_incoming_damage = 23
    modify_incoming_heal = 24
    modify_incoming_heal_flat = 116
    modify_incoming_damage_type = 25
    modify_incoming_armor_piercing = 26
    modify_outgoing_damage = 27
    modify_outgoing_damage_flat = 119
    modify_outgoing_heal = 28
    modify_outgoing_heal_flat = 118
    modify_outgoing_damage_type = 29
    modify_outgoing_armor_piercing = 30
    bounce_next = 31
    bounce_previous = 32
    bounce_back = 33
    bounce_all = 34
    absorb_damage = 35
    absorb_heal = 36
    modify_accuracy = 37
    dispel = 38
    confusion = 39
    cloaked_charm = 40
    cloaked_ward = 41
    stun_resist = 42
    clue = 109
    pip_conversion = 43
    crit_boost = 44
    crit_block = 45
    polymorph = 46
    delay_cast = 47
    modify_card_cloak = 48
    modify_card_damage = 49
    modify_card_accuracy = 51
    modify_card_mutation = 52
    modify_card_rank = 53
    modify_card_armor_piercing = 54
    summon_creature = 63
    teleport_player = 64
    stun = 65
    dampen = 66
    reshuffle = 67
    mind_control = 68
    modify_pips = 69
    modify_power_pips = 70
    modify_shadow_pips = 71
    modify_hate = 72
    damage_over_time = 73
    heal_over_time = 74
    modify_power_pip_chance = 75
    modify_rank = 76
    stun_block = 77
    reveal_cloak = 78
    instant_kill = 79
    after_life = 80
    deferred_damage = 81
    damage_per_total_pip_power = 82
    modify_card_heal = 50
    modify_card_charm = 55
    modify_card_warn = 56
    modify_card_outgoing_damage = 57
    modify_card_outgoing_accuracy = 58
    modify_card_outgoing_heal = 59
    modify_card_outgoing_armor_piercing = 60
    modify_card_incoming_damage = 61
    modify_card_absorb_damage = 62
    cloaked_ward_no_remove = 84
    add_combat_trigger_list = 85
    remove_combat_trigger_list = 86
    backlash_damage = 87
    modify_backlash = 88
    intercept = 89
    shadow_self = 90
    shadow_creature = 91
    modify_shadow_creature_level = 92
    select_shadow_creature_attack_target = 93
    shadow_decrement_turn = 94
    crit_boost_school_specific = 95
    spawn_creature = 96
    unpolymorph = 97
    power_pip_conversion = 98
    protect_card_beneficial = 99
    protect_card_harmful = 100
    protect_beneficial = 101
    protect_harmful = 102
    divide_damage = 103
    collect_essence = 104
    kill_creature = 105
    dispel_block = 106
    confusion_block = 107
    modify_pip_round_rate = 108
    max_health_damage = 110
    untargetable = 112
    make_targetable = 113
    force_targetable = 114
    remove_stun_block = 115
    exit_combat = 120
    suspend_pips = 121
    resume_pips = 122
    auto_pass = 123
    stop_auto_pass = 124
    vanish = 125
    stop_vanish = 126
    max_health_heal = 127
    heal_by_ward = 128
    taunt = 129
    pacify = 130
    remove_target_restriction = 131
    convert_hanging_effect = 132
    add_spell_to_deck = 133
    add_spell_to_hand = 134
    modify_incoming_damage_over_time = 135
    modify_incoming_heal_over_time = 136
    modify_card_damage_by_rank = 137


class EffectTarget(Enum):
    invalid_target = 0
    spell = 1
    specific_spells = 2
    target_global = 3
    enemy_team = 4
    enemy_team_all_at_once = 5
    friendly_team = 6
    friendly_team_all_at_once = 7
    enemy_single = 8
    friendly_single = 9
    minion = 10
    friendly_minion = 17
    self = 11
    at_least_one_enemy = 13
    preselected_enemy_single = 12
    multi_target_enemy = 14
    multi_target_friendly = 15
    friendly_single_not_me = 16


class ObjectType(Enum):
    undefined = 0
    player = 1
    npc = 2
    prop = 3
    object = 4
    house = 5
    key = 6
    old_key = 7
    deed = 8
    mail = 9
    recipe = 17
    equip_head = 10
    equip_chest = 11
    equip_legs = 12
    equip_hands = 13
    equip_finger = 14
    equip_feet = 15
    equip_ear = 16
    building_block = 18
    building_block_solid = 19
    golf = 20
    door = 21
    pet = 22
    fabric = 23
    window = 24
    roof = 25
    horse = 26
    structure = 27
    housing_texture = 28
    plant = 29


# class MagicSchool(Enum):
#     ice = 72777
#     sun = 78483
#     life = 2330892
#     fire = 2343174
#     star = 2625203
#     myth = 2448141
#     moon = 2504141
#     death = 78318724
#     storm = 83375795
#     gardening = 663550619
#     castle_magic = 806477568
#     whirly_burly = 931528087
#     balance = 1027491821
#     shadow = 1429009101
#     fishing = 1488274711
#     cantrips = 1760873841



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wizwalker',
 'wizwalker.cli',
 'wizwalker.combat',
 'wizwalker.extensions.scripting',
 'wizwalker.file_readers',
 'wizwalker.memory',
 'wizwalker.memory.memory_objects']

package_data = \
{'': ['*']}

install_requires = \
['aioconsole>=0.2.1,<0.3.0',
 'aiofiles>=0.5.0,<0.6.0',
 'aiomonitor>=0.4.5,<0.5.0',
 'appdirs>=1.4.4,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'click_default_group>=1.2.2,<2.0.0',
 'janus>=0.6.1,<0.7.0',
 'lark>=0.11.3,<0.12.0',
 'loguru>=0.5.1,<0.6.0',
 'pymem==1.8.3',
 'terminaltables>=3.1.0,<4.0.0']

entry_points = \
{'console_scripts': ['wizwalker = wizwalker.__main__:main']}

setup_kwargs = {
    'name': 'wizwalker',
    'version': '1.2.2',
    'description': 'Wizard101 scripting library',
    'long_description': "# WizWalker\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nWizard101 scripting library\n\n## documentation\nyou can find the documentation [here](https://starrfox.github.io/wizwalker/)\n\n## install\ndownload the latest release from [here](https://github.com/StarrFox/WizWalker/releases)\nor install from pypi `pip install -U wizwalker`\n\n## discord\njoin the offical discord [here](https://discord.gg/JHrdCNK)\n\n## development install\nThis package uses [poetry](https://python-poetry.org/)\n```shell script\n$ poetry install\n```\n\n## running\nShell may need admin perms\n```shell script\n$ poetry shell\n$ wizwalker\n```\n\n## building\nYou'll need the dev install (see above) for this to work\n\n### exe\n```shell script\n# Admin if needed\n$ pyinstaller -F --uac-admin --name WizWalker wizwalker/__main__.py\n# Normal\n$ pyinstaller -F --name WizWalker wizwalker/__main__.py\n```\n\n### wheel and source\n```shell script\n$ poetry build\n```\n\n### Docs\n```shell script\n$ cd docs\n$ make html\n```\n\n## console commands\nwizwalker: Runs the wizwalker cli\n\nwizwalker start-wiz: start wizard101 instances\n\nwizwalker wad: edit and extract wizard101 wad files\n",
    'author': 'StarrFox',
    'author_email': 'starrfox6312@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/StarrFox/wizwalker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from wizwalker.memory.memory_object import PropertyClass, DynamicMemoryObject


class ClientZone(PropertyClass):
    """
    Base class for ClientZones
    """

    async def read_base_address(self) -> int:
        raise NotImplementedError()

    async def zone_id(self) -> int:
        """
        This client zone's zone id
        """
        return await self.read_value_from_offset(72, "long long")

    async def write_zone_id(self, zone_id: int):
        """
        Write this client zone's zone id

        Args:
            zone_id: The zone id to write
        """
        await self.write_value_to_offset(72, zone_id, "long long")

    async def zone_name(self) -> str:
        """
        This client zone's zone name
        """
        return await self.read_string_from_offset(88)

    async def write_zone_name(self, zone_name: str):
        """
        Write this client zone's zone name

        Args:
            zone_name: The zone name to write
        """
        await self.write_string_to_offset(88, zone_name)


class DynamicClientZone(DynamicMemoryObject, ClientZone):
    """
    Dynamic client zone that can take an address
    """

    pass



from warnings import warn

import wizwalker


class CombatMember:
    def __init__(
        self,
        combat_handler: "wizwalker.combat.CombatHandler",
        combatant_control: "wizwalker.memory.DynamicWindow",
    ):
        self.combat_handler = combat_handler

        self._combatant_control = combatant_control

    # TODO: remove in 2.0
    async def get_particpant(self):
        warn(
            "get_particpant will be removed in 2.0 please use get_participant instead",
            DeprecationWarning,
        )
        return await self.get_participant()

    async def get_participant(self):
        """
        Get the underlying participant object
        """
        return await self._combatant_control.maybe_combat_participant()

    async def get_stats(self):
        """
        Get the underlying game stats object
        """
        part = await self.get_participant()
        return await part.game_stats()

    async def get_health_text_window(self) -> "wizwalker.memory.DynamicWindow":
        """
        Get the health text window
        Useful for targeting
        """
        possible = await self._combatant_control.get_windows_with_name("Health")
        if possible:
            return possible[0]

        raise ValueError("Couldn't find health child")

    async def get_name_text_window(self) -> "wizwalker.memory.DynamicWindow":
        """
        Get the name text window
        """
        possible = await self._combatant_control.get_windows_with_name("Name")
        if possible:
            return possible[0]

        raise ValueError("Couldn't find name child")

    async def is_client(self) -> bool:
        """
        If this member is the local client
        """
        owner_id = await self.owner_id()
        character_id = await self.combat_handler.client.client_object.character_id()
        return owner_id - 2 == character_id

    async def is_player(self) -> bool:
        """
        If this member is a player
        """
        part = await self.get_participant()
        return await part.is_player()

    async def is_monster(self) -> bool:
        """
        If this member is not a player and not a minion
        """
        return not await self.is_player() and not await self.is_minion()

    async def is_minion(self) -> bool:
        """
        If this member is a minion
        """
        part = await self.get_participant()
        return await part.is_minion()

    async def is_boss(self) -> bool:
        """
        If this member is a boss
        """
        part = await self.get_participant()
        return await part.boss_mob()

    async def is_stunned(self) -> bool:
        """
        If this member is stunned
        """
        part = await self.get_participant()
        return await part.stunned() != 0

    async def name(self) -> str:
        """
        Name of this member
        """
        name_window = await self.get_name_text_window()
        return await name_window.maybe_text()

    # TODO: finish
    # async def school_name(self) -> str:
    #     pass

    async def owner_id(self) -> int:
        """
        This member's owner id
        """
        part = await self.get_participant()
        return await part.owner_id_full()

    async def template_id(self) -> int:
        """
        This member's template id
        """
        part = await self.get_participant()
        return await part.template_id_full()

    async def normal_pips(self) -> int:
        """
        The number of normal pips this member has
        """
        part = await self.get_participant()
        return await part.num_pips()

    async def power_pips(self) -> int:
        """
        The number of power pips this member has
        """
        part = await self.get_participant()
        return await part.num_power_pips()

    async def shadow_pips(self) -> int:
        """
        The number of shadow pips this member has
        """
        part = await self.get_participant()
        return await part.num_shadow_pips()

    async def health(self) -> int:
        """
        The amount of health this member has
        """
        stats = await self.get_stats()
        return await stats.current_hitpoints()

    async def max_health(self) -> int:
        """
        This member's max health
        """
        stats = await self.get_stats()
        return await stats.max_hitpoints()

    async def mana(self) -> int:
        """
        The amount of mana this member has
        """
        stats = await self.get_stats()
        return await stats.current_mana()

    async def max_mana(self) -> int:
        """
        This member's max mana
        """
        stats = await self.get_stats()
        return await stats.max_mana()

    async def level(self) -> int:
        """
        This member's level
        """
        stats = await self.get_stats()
        return await stats.reference_level()



from wizwalker import XYZ
from wizwalker.memory.memory_object import MemoryObject


class CurrentQuestPosition(MemoryObject):
    async def read_base_address(self) -> int:
        return await self.hook_handler.read_current_quest_base()

    async def position(self) -> XYZ:
        return await self.read_xyz(0)

    async def write_position(self, position: XYZ):
        await self.write_xyz(0, position)



from typing import List, Optional

from wizwalker.memory.memory_object import DynamicMemoryObject, PropertyClass
from .enums import PipAquiredByEnum
from .game_stats import DynamicGameStats
from .spell import DynamicHand
from .play_deck import DynamicPlayDeck
from .spell_effect import DynamicSpellEffect


class CombatParticipant(PropertyClass):
    """
    Base class for CombatParticipants
    """

    def read_base_address(self) -> int:
        raise NotImplementedError()

    async def owner_id_full(self) -> int:
        """
        This combat participant's owner id
        """
        return await self.read_value_from_offset(112, "unsigned long long")

    async def write_owner_id_full(self, owner_id_full: int):
        """
        Write this combat participant's owner id

        Args:
            owner_id_full: The owner id to write
        """
        await self.write_value_to_offset(112, owner_id_full, "unsigned long long")

    async def template_id_full(self) -> int:
        """
        This combat participant's template id
        """
        return await self.read_value_from_offset(120, "unsigned long long")

    async def write_template_id_full(self, template_id_full: int):
        """
        Write this combat participant's template id

        Args:
            template_id_full: The template id to write
        """
        await self.write_value_to_offset(120, template_id_full, "unsigned long long")

    async def is_player(self) -> bool:
        """
        If this combat participant is a player
        """
        return await self.read_value_from_offset(128, "bool")

    async def write_is_player(self, is_player: bool):
        """
        Write if this combat participant is a player

        Args:
            is_player: The bool to write
        """
        await self.write_value_to_offset(128, is_player, "bool")

    async def zone_id_full(self) -> int:
        """
        This combat participant's zone id
        """
        return await self.read_value_from_offset(136, "unsigned long long")

    async def write_zone_id_full(self, zone_id_full: int):
        """
        Write this combat participant's zone id

        Args:
            zone_id_full: The zone id to write
        """
        await self.write_value_to_offset(136, zone_id_full, "unsigned long long")

    # TODO: look into what a team id is; i.e is it always the two ids
    async def team_id(self) -> int:
        """
        This combat participant's team id
        """
        return await self.read_value_from_offset(144, "int")

    async def write_team_id(self, team_id: int):
        """
        Write this combat participant's team id

        Args:
            team_id: The team id to write
        """
        await self.write_value_to_offset(144, team_id, "int")

    # TODO: turn this into an enum?
    async def primary_magic_school_id(self) -> int:
        """
        This combat participant's primary school id

        Notes:
            This is a template id
        """
        return await self.read_value_from_offset(148, "int")

    async def write_primary_magic_school_id(self, primary_magic_school_id: int):
        """
        Write this combat participant's primate school id

        Args:
            primary_magic_school_id: The school id to write

        Notes:
            this is a template id
        """
        await self.write_value_to_offset(148, primary_magic_school_id, "int")

    async def num_pips(self) -> int:
        """
        The number of pips this combat participant has
        """
        return await self.read_value_from_offset(152, "unsigned char")

    async def write_num_pips(self, num_pips: int):
        """
        Write this participant's pip number

        Args:
            num_pips: The pip number to write
        """
        await self.write_value_to_offset(152, num_pips, "unsigned char")

    async def num_power_pips(self) -> int:
        """
        The number of power pips this combat participant has
        """
        return await self.read_value_from_offset(153, "unsigned char")

    async def write_num_power_pips(self, num_power_pips: int):
        """
        Write the number of power pips this combat participant has

        Args:
            num_power_pips: The power pip number to write
        """
        await self.write_value_to_offset(153, num_power_pips, "unsigned char")

    async def num_shadow_pips(self) -> int:
        """
        The number of shadow pips this combat participant has
        """
        return await self.read_value_from_offset(154, "unsigned char")

    async def write_num_shadow_pips(self, num_shadow_pips: int):
        """
        Write the number of shadow pips this combat participant has

        Args:
            num_shadow_pips: The power pip number to write
        """
        await self.write_value_to_offset(154, num_shadow_pips, "unsigned char")

    # async def pip_round_rates(self) -> class SharedPointer<class ModifyPipRoundRateData>:
    #     return await self.read_value_from_offset(160, "class SharedPointer<class ModifyPipRoundRateData>")

    async def pips_suspended(self) -> bool:
        """
        If this participant's pips are suspended
        """
        return await self.read_value_from_offset(176, "bool")

    async def write_pips_suspended(self, pips_suspended: bool):
        """
        Write if this participant's pips are suspended

        Args:
            pips_suspended: bool if pips are suspended
        """
        await self.write_value_to_offset(176, pips_suspended, "bool")

    # TODO: finish docs
    async def stunned(self) -> int:
        return await self.read_value_from_offset(180, "int")

    async def write_stunned(self, stunned: int):
        await self.write_value_to_offset(180, stunned, "int")

    async def mindcontrolled(self) -> int:
        return await self.read_value_from_offset(208, "int")

    async def write_mindcontrolled(self, mindcontrolled: int):
        await self.write_value_to_offset(208, mindcontrolled, "int")

    async def original_team(self) -> int:
        return await self.read_value_from_offset(216, "int")

    async def write_original_team(self, original_team: int):
        await self.write_value_to_offset(216, original_team, "int")

    async def aura_turn_length(self) -> int:
        return await self.read_value_from_offset(228, "int")

    async def write_aura_turn_length(self, n_aura_turn_length: int):
        await self.write_value_to_offset(228, n_aura_turn_length, "int")

    async def clue(self) -> int:
        return await self.read_value_from_offset(220, "int")

    async def write_clue(self, clue: int):
        await self.write_value_to_offset(220, clue, "int")

    async def rounds_dead(self) -> int:
        return await self.read_value_from_offset(224, "int")

    async def write_rounds_dead(self, rounds_dead: int):
        await self.write_value_to_offset(224, rounds_dead, "int")

    async def polymorph_turn_length(self) -> int:
        return await self.read_value_from_offset(232, "int")

    async def write_polymorph_turn_length(self, n_polymorph_turn_length: int):
        await self.write_value_to_offset(232, n_polymorph_turn_length, "int")

    async def player_health(self) -> int:
        return await self.read_value_from_offset(236, "int")

    async def write_player_health(self, player_health: int):
        await self.write_value_to_offset(236, player_health, "int")

    async def max_player_health(self) -> int:
        return await self.read_value_from_offset(240, "int")

    async def write_max_player_health(self, max_player_health: int):
        await self.write_value_to_offset(240, max_player_health, "int")

    async def hide_current_hp(self) -> bool:
        return await self.read_value_from_offset(244, "bool")

    async def write_hide_current_hp(self, _hide_current_h_p: bool):
        await self.write_value_to_offset(244, _hide_current_h_p, "bool")

    async def max_hand_size(self) -> int:
        return await self.read_value_from_offset(248, "int")

    async def write_max_hand_size(self, max_hand_size: int):
        await self.write_value_to_offset(248, max_hand_size, "int")

    async def hand(self) -> DynamicHand:
        addr = await self.read_value_from_offset(256, "long long")
        return DynamicHand(self.hook_handler, addr)

    async def saved_hand(self) -> DynamicHand:
        addr = await self.read_value_from_offset(264, "long long")
        return DynamicHand(self.hook_handler, addr)

    async def play_deck(self) -> Optional[DynamicPlayDeck]:
        addr = await self.read_value_from_offset(272, "long long")

        if addr == 0:
            return None

        return DynamicPlayDeck(self.hook_handler, addr)

    async def saved_play_deck(self) -> Optional[DynamicPlayDeck]:
        addr = await self.read_value_from_offset(280, "long long")

        if addr == 0:
            return None

        return DynamicPlayDeck(self.hook_handler, addr)

    async def saved_game_stats(self) -> DynamicGameStats:
        addr = await self.read_value_from_offset(288, "long long")
        return DynamicGameStats(self.hook_handler, addr)

    async def saved_primary_magic_school_id(self) -> int:
        return await self.read_value_from_offset(304, "int")

    async def write_saved_primary_magic_school_id(
        self, saved_primary_magic_school_id: int
    ):
        await self.write_value_to_offset(304, saved_primary_magic_school_id, "int")

    async def game_stats(self) -> DynamicGameStats:
        addr = await self.read_value_from_offset(312, "long long")
        return DynamicGameStats(self.hook_handler, addr)

    # TODO: figure out what color is
    # async def color(self) -> class Color:
    #     return await self.read_value_from_offset(328, "class Color")
    #
    # async def write_color(self, color: class Color):
    #     await self.write_value_to_offset(328, color, "class Color")

    async def rotation(self) -> float:
        return await self.read_value_from_offset(332, "float")

    async def write_rotation(self, rotation: float):
        await self.write_value_to_offset(332, rotation, "float")

    async def radius(self) -> float:
        return await self.read_value_from_offset(336, "float")

    async def write_radius(self, radius: float):
        await self.write_value_to_offset(336, radius, "float")

    async def subcircle(self) -> int:
        return await self.read_value_from_offset(340, "int")

    async def write_subcircle(self, subcircle: int):
        await self.write_value_to_offset(340, subcircle, "int")

    async def pvp(self) -> bool:
        return await self.read_value_from_offset(344, "bool")

    async def write_pvp(self, pvp: bool):
        await self.write_value_to_offset(344, pvp, "bool")

    async def accuracy_bonus(self) -> float:
        return await self.read_value_from_offset(388, "float")

    async def write_accuracy_bonus(self, accuracy_bonus: float):
        await self.write_value_to_offset(388, accuracy_bonus, "float")

    async def minion_sub_circle(self) -> int:
        return await self.read_value_from_offset(392, "int")

    async def write_minion_sub_circle(self, minion_sub_circle: int):
        await self.write_value_to_offset(392, minion_sub_circle, "int")

    async def is_minion(self) -> bool:
        return await self.read_value_from_offset(396, "bool")

    async def write_is_minion(self, is_minion: bool):
        await self.write_value_to_offset(396, is_minion, "bool")

    async def hanging_effects(self) -> List[DynamicSpellEffect]:
        hanging_effects = []
        for addr in await self.read_linked_list(408):
            hanging_effects.append(DynamicSpellEffect(self.hook_handler, addr))

        return hanging_effects

    async def public_hanging_effects(self) -> List[DynamicSpellEffect]:
        hanging_effects = []
        for addr in await self.read_linked_list(424):
            hanging_effects.append(DynamicSpellEffect(self.hook_handler, addr))

        return hanging_effects

    async def aura_effects(self) -> List[DynamicSpellEffect]:
        aura_effects = []
        for addr in await self.read_linked_list(440):
            aura_effects.append(DynamicSpellEffect(self.hook_handler, addr))

        return aura_effects

    # TODO: add this class
    # async def shadow_effects(self) -> class SharedPointer<class ShadowSpellTrackingData>:
    #     return await self.read_value_from_offset(456, "class SharedPointer<class ShadowSpellTrackingData>")

    async def shadow_spell_effects(self) -> List[DynamicSpellEffect]:
        shadow_spell_effects = []
        for addr in await self.read_linked_list(472):
            shadow_spell_effects.append(DynamicSpellEffect(self.hook_handler, addr))

        return shadow_spell_effects

    async def death_activated_effects(self) -> List[DynamicSpellEffect]:
        death_activated_effects = []
        for addr in await self.read_shared_linked_list(504):
            death_activated_effects.append(DynamicSpellEffect(self.hook_handler, addr))

        return death_activated_effects

    # note: these are actually DelaySpellEffects
    async def delay_cast_effects(self) -> List[DynamicSpellEffect]:
        delay_cast_effects = []
        for addr in await self.read_linked_list(520):
            delay_cast_effects.append(DynamicSpellEffect(self.hook_handler, addr))

        return delay_cast_effects

    async def polymorph_spell_template_id(self) -> int:
        return await self.read_value_from_offset(568, "unsigned int")

    async def write_polymorph_spell_template_id(self, polymorph_spell_template_id: int):
        await self.write_value_to_offset(
            568, polymorph_spell_template_id, "unsigned int"
        )

    async def side(self) -> str:
        return await self.read_string_from_offset(592)

    async def write_side(self, side: str):
        await self.write_string_to_offset(592, side)

    async def shadow_spells_disabled(self) -> bool:
        return await self.read_value_from_offset(637, "bool")

    async def write_shadow_spells_disabled(self, shadow_spells_disabled: bool):
        await self.write_value_to_offset(637, shadow_spells_disabled, "bool")

    async def boss_mob(self) -> bool:
        return await self.read_value_from_offset(638, "bool")

    async def write_boss_mob(self, boss_mob: bool):
        await self.write_value_to_offset(638, boss_mob, "bool")

    async def hide_pvp_enemy_chat(self) -> bool:
        return await self.read_value_from_offset(639, "bool")

    async def write_hide_pvp_enemy_chat(self, hide_pvp_enemy_chat: bool):
        await self.write_value_to_offset(639, hide_pvp_enemy_chat, "bool")

    async def combat_trigger_ids(self) -> int:
        return await self.read_value_from_offset(664, "int")

    async def write_combat_trigger_ids(self, combat_trigger_ids: int):
        await self.write_value_to_offset(664, combat_trigger_ids, "int")

    async def backlash(self) -> int:
        return await self.read_value_from_offset(692, "int")

    async def write_backlash(self, backlash: int):
        await self.write_value_to_offset(692, backlash, "int")

    async def past_backlash(self) -> int:
        return await self.read_value_from_offset(696, "int")

    async def write_past_backlash(self, past_backlash: int):
        await self.write_value_to_offset(696, past_backlash, "int")

    async def shadow_creature_level(self) -> int:
        return await self.read_value_from_offset(700, "int")

    async def write_shadow_creature_level(self, shadow_creature_level: int):
        await self.write_value_to_offset(700, shadow_creature_level, "int")

    async def past_shadow_creature_level(self) -> int:
        return await self.read_value_from_offset(704, "int")

    async def write_past_shadow_creature_level(self, past_shadow_creature_level: int):
        await self.write_value_to_offset(704, past_shadow_creature_level, "int")

    async def shadow_creature_level_count(self) -> int:
        return await self.read_value_from_offset(712, "int")

    async def write_shadow_creature_level_count(self, shadow_creature_level_count: int):
        await self.write_value_to_offset(712, shadow_creature_level_count, "int")

    async def intercept_effect(self) -> Optional[DynamicSpellEffect]:
        addr = await self.read_value_from_offset(736, "long long")

        if addr == 0:
            return None

        return DynamicSpellEffect(self.hook_handler, addr)

    async def rounds_since_shadow_pip(self) -> int:
        return await self.read_value_from_offset(768, "int")

    async def write_rounds_since_shadow_pip(self, rounds_since_shadow_pip: int):
        await self.write_value_to_offset(768, rounds_since_shadow_pip, "int")

    async def polymorph_effect(self) -> Optional[DynamicSpellEffect]:
        addr = await self.read_value_from_offset(792, "long long")

        if addr == 0:
            return None

        return DynamicSpellEffect(self.hook_handler, addr)

    async def confused(self) -> int:
        return await self.read_value_from_offset(188, "int")

    async def write_confused(self, confused: int):
        await self.write_value_to_offset(188, confused, "int")

    async def confusion_trigger(self) -> int:
        return await self.read_value_from_offset(192, "int")

    async def write_confusion_trigger(self, confusion_trigger: int):
        await self.write_value_to_offset(192, confusion_trigger, "int")

    async def confusion_display(self) -> bool:
        return await self.read_value_from_offset(196, "bool")

    async def write_confusion_display(self, confusion_display: bool):
        await self.write_value_to_offset(196, confusion_display, "bool")

    async def confused_target(self) -> bool:
        return await self.read_value_from_offset(197, "bool")

    async def write_confused_target(self, confused_target: bool):
        await self.write_value_to_offset(197, confused_target, "bool")

    async def untargetable(self) -> bool:
        return await self.read_value_from_offset(198, "bool")

    async def write_untargetable(self, untargetable: bool):
        await self.write_value_to_offset(198, untargetable, "bool")

    async def untargetable_rounds(self) -> int:
        return await self.read_value_from_offset(200, "int")

    async def write_untargetable_rounds(self, untargetable_rounds: int):
        await self.write_value_to_offset(200, untargetable_rounds, "int")

    async def restricted_target(self) -> bool:
        return await self.read_value_from_offset(204, "bool")

    async def write_restricted_target(self, restricted_target: bool):
        await self.write_value_to_offset(204, restricted_target, "bool")

    async def exit_combat(self) -> bool:
        return await self.read_value_from_offset(205, "bool")

    async def write_exit_combat(self, exit_combat: bool):
        await self.write_value_to_offset(205, exit_combat, "bool")

    async def stunned_display(self) -> bool:
        return await self.read_value_from_offset(184, "bool")

    async def write_stunned_display(self, stunned_display: bool):
        await self.write_value_to_offset(184, stunned_display, "bool")

    async def mindcontrolled_display(self) -> bool:
        return await self.read_value_from_offset(212, "bool")

    async def write_mindcontrolled_display(self, mindcontrolled_display: bool):
        await self.write_value_to_offset(212, mindcontrolled_display, "bool")

    async def auto_pass(self) -> bool:
        return await self.read_value_from_offset(688, "bool")

    async def write_auto_pass(self, auto_pass: bool):
        await self.write_value_to_offset(688, auto_pass, "bool")

    async def vanish(self) -> bool:
        return await self.read_value_from_offset(689, "bool")

    async def write_vanish(self, vanish: bool):
        await self.write_value_to_offset(689, vanish, "bool")

    async def my_team_turn(self) -> bool:
        return await self.read_value_from_offset(690, "bool")

    async def write_my_team_turn(self, my_team_turn: bool):
        await self.write_value_to_offset(690, my_team_turn, "bool")

    async def planning_phase_pip_aquired_type(self) -> PipAquiredByEnum:
        return await self.read_enum(784, PipAquiredByEnum)

    async def write_planning_phase_pip_aquired_type(
        self, planning_phase_pip_aquired_type: PipAquiredByEnum
    ):
        await self.write_enum(784, planning_phase_pip_aquired_type)

    # async def cheat_settings(self) -> class SharedPointer<class CombatCheatSettings>:
    #     return await self.read_value_from_offset(96, "class SharedPointer<class CombatCheatSettings>")

    async def is_monster(self) -> int:
        return await self.read_value_from_offset(400, "unsigned int")

    async def write_is_monster(self, is_monster: int):
        await self.write_value_to_offset(400, is_monster, "unsigned int")

    # async def weapon_nif_sound_list(self) -> class SharedPointer<class SpellNifSoundOverride>:
    #     return await self.read_value_from_offset(80, "class SharedPointer<class SpellNifSoundOverride>")

    async def pet_combat_trigger(self) -> int:
        return await self.read_value_from_offset(680, "int")

    async def write_pet_combat_trigger(self, pet_combat_trigger: int):
        await self.write_value_to_offset(680, pet_combat_trigger, "int")

    async def pet_combat_trigger_target(self) -> int:
        return await self.read_value_from_offset(684, "int")

    async def write_pet_combat_trigger_target(self, pet_combat_trigger_target: int):
        await self.write_value_to_offset(684, pet_combat_trigger_target, "int")

    async def shadow_pip_rate_threshold(self) -> float:
        return await self.read_value_from_offset(808, "float")

    async def write_shadow_pip_rate_threshold(self, shadow_pip_rate_threshold: float):
        await self.write_value_to_offset(808, shadow_pip_rate_threshold, "float")

    async def base_spell_damage(self) -> int:
        return await self.read_value_from_offset(812, "int")

    async def write_base_spell_damage(self, base_spell_damage: int):
        await self.write_value_to_offset(812, base_spell_damage, "int")

    async def stat_damage(self) -> float:
        return await self.read_value_from_offset(816, "float")

    async def write_stat_damage(self, stat_damage: float):
        await self.write_value_to_offset(816, stat_damage, "float")

    async def stat_resist(self) -> float:
        return await self.read_value_from_offset(820, "float")

    async def write_stat_resist(self, stat_resist: float):
        await self.write_value_to_offset(820, stat_resist, "float")

    async def stat_pierce(self) -> float:
        return await self.read_value_from_offset(824, "float")

    async def write_stat_pierce(self, stat_pierce: float):
        await self.write_value_to_offset(824, stat_pierce, "float")

    async def mob_level(self) -> int:
        return await self.read_value_from_offset(828, "int")

    async def write_mob_level(self, mob_level: int):
        await self.write_value_to_offset(828, mob_level, "int")


class DynamicCombatParticipant(DynamicMemoryObject, CombatParticipant):
    pass



import asyncio
from functools import cached_property
from typing import Callable, List, Optional

import pymem

from . import (
    CacheHandler,
    Keycode,
    MemoryReadError,
    ReadingEnumFailed,
    utils,
)
from .constants import WIZARD_SPEED
from .memory import (
    CurrentActorBody,
    CurrentClientObject,
    CurrentDuel,
    CurrentGameStats,
    CurrentQuestPosition,
    CurrentRootWindow,
    DuelPhase,
    HookHandler,
    CurrentRenderContext,
)
from .mouse_handler import MouseHandler
from .utils import (
    XYZ,
    check_if_process_running,
    get_window_title,
    set_window_title,
    get_window_rectangle,
)


class Client:
    """
    Represents a connected wizard client

    Args:
        window_handle: A handle to the window this client connects to
    """

    def __init__(self, window_handle: int):
        self.window_handle = window_handle

        self._pymem = pymem.Pymem()
        self._pymem.open_process_from_id(self.process_id)
        self.hook_handler = HookHandler(self._pymem, self)

        self.cache_handler = CacheHandler()
        self.mouse_handler = MouseHandler(self)

        self.stats = CurrentGameStats(self.hook_handler)
        self.body = CurrentActorBody(self.hook_handler)
        self.duel = CurrentDuel(self.hook_handler)
        self.quest_position = CurrentQuestPosition(self.hook_handler)
        self.client_object = CurrentClientObject(self.hook_handler)
        self.root_window = CurrentRootWindow(self.hook_handler)
        self.render_context = CurrentRenderContext(self.hook_handler)

        self._template_ids = None
        self._is_loading_addr = None
        self._world_view_window = None

    def __repr__(self):
        return f"<Client {self.window_handle=} {self.process_id=}>"

    @property
    def title(self):
        """
        Get or set this window's title
        """
        return get_window_title(self.window_handle)

    @title.setter
    def title(self, window_title: str):
        set_window_title(self.window_handle, window_title)

    @property
    def window_rectangle(self):
        """
        Get this client's window rectangle
        """
        return get_window_rectangle(self.window_handle)

    @cached_property
    def process_id(self) -> int:
        """
        Client's process id
        """
        return utils.get_pid_from_handle(self.window_handle)

    def is_running(self):
        """
        If this client is still running
        """
        return check_if_process_running(self._pymem.process_handle)

    async def zone_name(self) -> Optional[str]:
        """
        Client's current zone name
        """
        client_zone = await self.client_object.client_zone()

        if client_zone is not None:
            return await client_zone.zone_name()

        return None

    async def get_base_entity_list(self):
        """
        List of WizClientObjects currently loaded
        """
        root_client = await self.client_object.parent()
        return await root_client.children()

    async def get_base_entities_with_name(self, name: str):
        """
        Get entites with a name

        Args:
            name: The name to search for

        Returns:
            List of the matching entities
        """

        async def _pred(entity):
            object_template = await entity.object_template()
            return await object_template.object_name() == name

        return await self.get_base_entities_with_predicate(_pred)

    # TODO: add example
    async def get_base_entities_with_predicate(self, predicate: Callable):
        """
        Get entities with a predicate

        Args:
            predicate: Awaitable that returns True or False on if to add an entity

        Returns:
            The matching entities
        """
        entities = []

        for entity in await self.get_base_entity_list():
            if await predicate(entity):
                entities.append(entity)

        return entities

    async def get_world_view_window(self):
        """
        Get the world view window
        """
        if self._world_view_window:
            return self._world_view_window

        pos = await self.root_window.get_windows_with_name("WorldView")
        # TODO: test this claim on login screen
        # world view always exists
        self._world_view_window = pos[0]
        return self._world_view_window

    async def activate_hooks(
        self, *, wait_for_ready: bool = True, timeout: float = None
    ):
        """
        Activate all memory hooks but mouseless

        Keyword Args:
            wait_for_ready: If this should wait for hooks to be ready to use (duel exempt)
            timeout: How long to wait for hook values to be witten (None for no timeout)
        """
        await self.hook_handler.activate_all_hooks(
            wait_for_ready=wait_for_ready, timeout=timeout
        )

    async def close(self):
        """
        Closes this client; unhooking all active hooks
        """
        # if the client isn't running there isn't anything to unhook
        if not self.is_running():
            return

        await self.hook_handler.close()

    async def get_template_ids(self) -> dict:
        """
        Get a dict of template ids mapped to their value
        ids are str
        """
        if self._template_ids:
            return self._template_ids

        self._template_ids = await self.cache_handler.get_template_ids()
        return self._template_ids

    async def in_battle(self) -> bool:
        """
        If the client is in battle or not
        """
        try:
            duel_phase = await self.duel.duel_phase()
        except (ReadingEnumFailed, MemoryReadError):
            return False
        else:
            return duel_phase is not DuelPhase.ended

    async def is_loading(self) -> bool:
        """
        If the client is currently in a loading screen
        (does not apply to character load in)
        """
        if not self._is_loading_addr:
            mov_instruction_addr = await self.hook_handler.pattern_scan(
                b"\xC6\x05....\x00\xC6\x80.....\x48\x8B",
                module="WizardGraphicalClient.exe",
            )
            # first 2 bytes are the mov instruction and mov type (C6 05)
            rip_offset = await self.hook_handler.read_typed(
                mov_instruction_addr + 2, "int"
            )
            # 7 is the length of this instruction
            self._is_loading_addr = mov_instruction_addr + 7 + rip_offset

        # 1 -> can't move (loading) 0 -> can move (not loading)
        return await self.hook_handler.read_typed(self._is_loading_addr, "bool")

    async def is_in_dialog(self) -> bool:
        """
        If the client is in dialog
        """
        world_view = await self.get_world_view_window()
        for child in await world_view.children():
            # TODO: check if we also need to check for wndDialogMain child
            if (child_name := await child.name()) == "NPCServicesWin":
                return True

            elif child_name == "wndDialogMain":
                return True

        return False

    async def is_in_npc_range(self) -> bool:
        """
        If the client is within an npc interact range
        """
        world_view = await self.get_world_view_window()
        for child in await world_view.children():
            if await child.name() == "NPCRangeWin":
                return True

        return False

    async def backpack_space(self) -> tuple:
        """
        This client's backpack space used and max
        must be on inventory page to use
        """
        maybe_space_window = await self.root_window.get_windows_with_name(
            "inventorySpace"
        )

        if not maybe_space_window:
            # TODO: replace error
            raise ValueError("must open inventory screen to get")

        text = await maybe_space_window[0].maybe_text()
        text = text.replace("<center>", "")
        used, total = text.split("/")
        return int(used), int(total)

    async def wait_for_zone_change(
        self, name: Optional[str] = None, *, sleep_time: Optional[float] = 0.5
    ):
        """
        Wait for the client's zone to change

        Args:
            name: The name of the zone to wait to be changed from or None to read
            sleep_time: How long to sleep between reads or None to not
        """
        if name is None:
            name = await self.zone_name()

            # the client zone is deallocated on zone change
            if name is None:
                raise RuntimeError("Client zone was already deallocated before calling")

        while (zone_name := await self.zone_name()) == name and zone_name is not None:
            await asyncio.sleep(sleep_time)

        while await self.is_loading():
            await asyncio.sleep(sleep_time)

    async def current_energy(self) -> int:
        """
        Client's current energy
        energy globe must be visible to use
        """
        maybe_energy_text = await self.root_window.get_windows_with_name("textEnergy")

        if not maybe_energy_text:
            # TODO: replace error
            raise ValueError("Energy globe not on screen")

        text = await maybe_energy_text[0].maybe_text()
        text = text.replace("<center>", "")
        text = text.replace("</center>", "")
        return int(text)

    def login(self, username: str, password: str):
        """
        Login this client

        Args:
            username: The username to login with
            password: The password to login with
        """
        utils.instance_login(self.window_handle, username, password)

    async def send_key(self, key: Keycode, seconds: float = 0):
        """
        Send a key

        Args:
            key: The Keycode to send
            seconds: How long to send it for
        """
        await utils.timed_send_key(self.window_handle, key, seconds)

    async def send_hotkey(self, modifers: List[Keycode], key: Keycode):
        """
        send a hotkey

        Args:
            modifers: The key modifers i.e CTRL, ALT
            key: The key being modified
        """
        await utils.send_hotkey(self.window_handle, modifers, key)

    async def goto(self, x: float, y: float):
        """
        Moves the player to a specific x and y

        Args:
            x: X to move to
            y: Y to move to
        """
        current_xyz = await self.body.position()
        # (40 / 100) + 1 = 1.4
        speed_multiplier = ((await self.client_object.speed_multiplier()) / 100) + 1
        target_xyz = utils.XYZ(x, y, current_xyz.z)
        distance = current_xyz - target_xyz
        move_seconds = distance / (WIZARD_SPEED * speed_multiplier)
        yaw = utils.calculate_perfect_yaw(current_xyz, target_xyz)

        await self.body.write_yaw(yaw)
        await utils.timed_send_key(self.window_handle, Keycode.W, move_seconds)

    async def teleport(self, xyz: XYZ, yaw: int = None):
        """
        Teleport the client

        Args:
            xyz: xyz to teleport to
            yaw: yaw to set or None to not change
        """
        await self.body.write_position(xyz)

        if yaw is not None:
            await self.body.write_yaw(yaw)



import asyncio
from typing import List, Optional, Union

import wizwalker


class CombatCard:
    """
    Represents a spell card
    """

    def __init__(
        self, combat_handler, spell_window: "wizwalker.memory.window.DynamicWindow",
    ):
        self.combat_handler = combat_handler

        self._spell_window = spell_window

    # TODO: add checks before casting
    async def cast(
        self,
        target: Union["CombatCard", "wizwalker.combat.CombatMember", None],
        *,
        sleep_time: Optional[float] = 1.0
    ):
        """
        Cast this Card on another Card; a Member or with no target

        Args:
            target: Card, Member, or None if there is no target
            sleep_time: how long to sleep after enchants and between multicasts or None for no sleep
        """
        if isinstance(target, CombatCard):
            cards_len_before = len(await self.combat_handler.get_cards())

            await self.combat_handler.client.mouse_handler.click_window(
                self._spell_window
            )
            await asyncio.sleep(sleep_time)
            await self.combat_handler.client.mouse_handler.click_window(
                target._spell_window
            )

            # wait until card number goes down
            while len(await self.combat_handler.get_cards()) > cards_len_before:
                await asyncio.sleep(0.1)

            # wiz can't keep up with how fast we can cast
            if sleep_time is not None:
                await asyncio.sleep(sleep_time)

        elif target is None:
            await self.combat_handler.client.mouse_handler.click_window(
                self._spell_window
            )
            # we don't need to sleep because nothing will be casted after

        else:
            await self.combat_handler.client.mouse_handler.click_window(
                self._spell_window
            )

            # see above
            if sleep_time is not None:
                await asyncio.sleep(sleep_time)

            await self.combat_handler.client.mouse_handler.click_window(
                await target.get_health_text_window()
            )

    async def discard(self, *, sleep_time: Optional[float] = 1.0):
        """
        Discard this Card

        Args:
            sleep_time: Time to sleep after discard or None to not
        """
        cards_len_before = len(await self.combat_handler.get_cards())
        await self.combat_handler.client.mouse_handler.click_window(
            self._spell_window, right_click=True
        )

        # wait until card number goes down
        while len(await self.combat_handler.get_cards()) > cards_len_before:
            await asyncio.sleep(0.1)

        if sleep_time is not None:
            await asyncio.sleep(sleep_time)

    async def get_graphical_spell(
        self,
    ) -> "wizwalker.memory.memory_objects.spell.DynamicGraphicalSpell":
        """
        The GraphicalSpell with information about this card
        """
        return await self._spell_window.maybe_graphical_spell()

    async def get_spell_effects(
        self,
    ) -> List["wizwalker.memory.memory_objects.spell_effect.DynamicSpellEffect"]:
        spell = await self.get_graphical_spell()
        return await spell.spell_effects()

    async def name(self) -> str:
        """
        The name of this card
        """
        graphical_spell = await self.get_graphical_spell()
        spell_template = await graphical_spell.spell_template()
        # name is the actual name; display name is some wack stuff
        return await spell_template.name()

    async def template_id(self) -> int:
        """
        This card's template id
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.template_id()

    async def spell_id(self) -> int:
        """
        This card's spell id
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.spell_id()

    async def accuracy(self) -> int:
        """
        Current accuracy of this card
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.accuracy()

    async def is_castable(self) -> bool:
        """
        If this card can be casted
        """
        spell_window = self._spell_window
        return not await spell_window.maybe_spell_grayed()

    async def is_enchanted(self) -> bool:
        """
        If this card is enchanted or not
        """
        grapical_spell = await self.get_graphical_spell()
        return await grapical_spell.enchantment() != 0

    async def is_treasure_card(self) -> bool:
        """
        If this card is a treasure card
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.treasure_card()

    async def is_item_card(self) -> bool:
        """
        If this card is an item card
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.item_card()

    async def is_side_board(self) -> bool:
        """
        If this card is from the side deck
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.side_board()

    async def is_cloaked(self) -> bool:
        """
        If this card is cloaked
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.cloaked()

    async def is_enchanted_from_item_card(self) -> bool:
        """
        If this card was enchanted from an item card
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.enchantment_spell_is_item_card()

    async def is_pve_only(self) -> bool:
        """
        If this card can only be used in pve
        """
        graphical_spell = await self.get_graphical_spell()
        return await graphical_spell.pve()



from typing import List

from wizwalker.memory.memory_object import PropertyClass, DynamicMemoryObject


class PlayDeck(PropertyClass):
    async def read_base_address(self) -> int:
        raise NotImplementedError()

    async def deck_to_save(self) -> List["DynamicPlaySpellData"]:
        spell_data = []
        for addr in await self.read_shared_vector(72):
            spell_data.append(DynamicPlaySpellData(self.hook_handler, addr))

        return spell_data

    async def graveyard_to_save(self) -> List["DynamicPlaySpellData"]:
        spell_data = []
        for addr in await self.read_shared_vector(96):
            spell_data.append(DynamicPlaySpellData(self.hook_handler, addr))

        return spell_data


class DynamicPlayDeck(DynamicMemoryObject, PlayDeck):
    pass


class PlaySpellData(PropertyClass):
    async def read_base_address(self) -> int:
        raise NotImplementedError()

    async def template_id(self) -> int:
        return await self.read_value_from_offset(72, "unsigned int")

    async def enchantment(self) -> int:
        return await self.read_value_from_offset(76, "unsigned int")


class DynamicPlaySpellData(DynamicMemoryObject, PlaySpellData):
    pass



from typing import List

from wizwalker.utils import XYZ
from wizwalker.memory.memory_object import PropertyClass
from .combat_participant import DynamicCombatParticipant
from .enums import DuelExecutionOrder, DuelPhase, SigilInitiativeSwitchMode
from .combat_resolver import DynamicCombatResolver


class Duel(PropertyClass):
    async def read_base_address(self) -> int:
        raise NotImplementedError()

    async def participant_list(self,) -> List[DynamicCombatParticipant]:
        pointers = await self.read_shared_vector(80)

        participants = []
        for addr in pointers:
            participants.append(DynamicCombatParticipant(self.hook_handler, addr))

        return participants

    async def duel_id_full(self) -> int:
        return await self.read_value_from_offset(72, "unsigned long long")

    async def write_duel_id_full(self, duel_id_full: int):
        await self.write_value_to_offset(72, duel_id_full, "unsigned long long")

    async def planning_timer(self) -> float:
        return await self.read_value_from_offset(112, "float")

    async def write_planning_timer(self, planning_timer: float):
        await self.write_value_to_offset(112, planning_timer, "float")

    async def position(self) -> XYZ:
        return await self.read_xyz(116)

    async def write_position(self, position: XYZ):
        await self.write_xyz(116, position)

    async def yaw(self) -> float:
        return await self.read_value_from_offset(128, "float")

    async def write_yaw(self, yaw: float):
        await self.write_value_to_offset(128, yaw, "float")

    async def disable_timer(self) -> bool:
        return await self.read_value_from_offset(146, "bool")

    async def write_disable_timer(self, disable_timer: bool):
        await self.write_value_to_offset(146, disable_timer, "bool")

    async def tutorial_mode(self) -> bool:
        return await self.read_value_from_offset(147, "bool")

    async def write_tutorial_mode(self, tutorial_mode: bool):
        await self.write_value_to_offset(147, tutorial_mode, "bool")

    async def first_team_to_act(self) -> int:
        return await self.read_value_from_offset(148, "int")

    async def write_first_team_to_act(self, first_team_to_act: int):
        await self.write_value_to_offset(148, first_team_to_act, "int")

    async def combat_resolver(self) -> DynamicCombatResolver:
        addr = await self.read_value_from_offset(104, "long long")
        return DynamicCombatResolver(self.hook_handler, addr)

    async def pvp(self) -> bool:
        return await self.read_value_from_offset(144, "bool")

    async def write_pvp(self, pvp: bool):
        await self.write_value_to_offset(144, pvp, "bool")

    async def battleground(self) -> bool:
        return await self.read_value_from_offset(145, "bool")

    async def write_battleground(self, b_battleground: bool):
        await self.write_value_to_offset(145, b_battleground, "bool")

    async def round_num(self) -> int:
        return await self.read_value_from_offset(156, "int")

    async def write_round_num(self, round_num: int):
        await self.write_value_to_offset(156, round_num, "int")

    async def execution_phase_timer(self) -> float:
        return await self.read_value_from_offset(164, "float")

    async def write_execution_phase_timer(self, execution_phase_timer: float):
        await self.write_value_to_offset(164, execution_phase_timer, "float")

    # note: this seems to be unused
    # async def execution_phase_combat_actions(self) -> class CombatAction:
    #     return await self.read_value_from_offset(168, "class CombatAction")

    # note: this also seems to be unused
    # async def sigil_actions(self) -> class CombatAction:
    #     return await self.read_value_from_offset(184, "class CombatAction")

    # async def shadow_pip_rule(self) -> class SharedPointer<class ShadowPipRule>:
    #     return await self.read_value_from_offset(240, "class SharedPointer<class ShadowPipRule>")

    # async def game_object_anim_state_tracker(self) -> class GameObjectAnimStateTracker:
    #     return await self.read_value_from_offset(256, "class GameObjectAnimStateTracker")

    async def duel_phase(self) -> DuelPhase:
        return await self.read_enum(160, DuelPhase)

    async def write_duel_phase(self, duel_phase: DuelPhase):
        await self.write_enum(160, duel_phase)

    # async def duel_modifier(self) -> class SharedPointer<class DuelModifier>:
    #     return await self.read_value_from_offset(224, "class SharedPointer<class DuelModifier>")

    async def initiative_switch_mode(self) -> SigilInitiativeSwitchMode:
        return await self.read_enum(344, SigilInitiativeSwitchMode)

    async def write_initiative_switch_mode(
        self, initiative_switch_mode: SigilInitiativeSwitchMode
    ):
        await self.write_enum(344, initiative_switch_mode)

    async def initiative_switch_rounds(self) -> int:
        return await self.read_value_from_offset(348, "int")

    async def write_initiative_switch_rounds(self, initiative_switch_rounds: int):
        await self.write_value_to_offset(348, initiative_switch_rounds, "int")

    # async def combat_rules(self) -> class SharedPointer<class CombatRule>:
    #     return await self.read_value_from_offset(424, "class SharedPointer<class CombatRule>")

    # async def alternate_turn_combat_rule(self) -> class SharedPointer<class AlternateTurnsCombatRule>:
    #     return await self.read_value_from_offset(440, "class SharedPointer<class AlternateTurnsCombatRule>")

    async def alt_turn_counter(self) -> int:
        return await self.read_value_from_offset(416, "int")

    async def write_alt_turn_counter(self, alt_turn_counter: int):
        await self.write_value_to_offset(416, alt_turn_counter, "int")

    async def original_first_team_to_act(self) -> int:
        return await self.read_value_from_offset(152, "int")

    async def write_original_first_team_to_act(self, original_first_team_to_act: int):
        await self.write_value_to_offset(152, original_first_team_to_act, "int")

    async def execution_order(self) -> DuelExecutionOrder:
        return await self.read_enum(456, DuelExecutionOrder)

    async def write_execution_order(self, execution_order: DuelExecutionOrder):
        await self.write_enum(456, execution_order)

    async def no_henchmen(self) -> bool:
        return await self.read_value_from_offset(460, "bool")

    async def write_no_henchmen(self, no_henchmen: bool):
        await self.write_value_to_offset(460, no_henchmen, "bool")

    async def spell_truncation(self) -> bool:
        return await self.read_value_from_offset(461, "bool")

    async def write_spell_truncation(self, spell_truncation: bool):
        await self.write_value_to_offset(461, spell_truncation, "bool")

    async def shadow_threshold_factor(self) -> float:
        return await self.read_value_from_offset(468, "float")

    async def write_shadow_threshold_factor(self, shadow_threshold_factor: float):
        await self.write_value_to_offset(468, shadow_threshold_factor, "float")

    async def shadow_pip_rating_factor(self) -> float:
        return await self.read_value_from_offset(472, "float")

    async def write_shadow_pip_rating_factor(self, shadow_pip_rating_factor: float):
        await self.write_value_to_offset(472, shadow_pip_rating_factor, "float")

    async def default_shadow_pip_rating(self) -> float:
        return await self.read_value_from_offset(476, "float")

    async def write_default_shadow_pip_rating(self, default_shadow_pip_rating: float):
        await self.write_value_to_offset(476, default_shadow_pip_rating, "float")

    async def shadow_pip_threshold_team0(self) -> float:
        return await self.read_value_from_offset(480, "float")

    async def write_shadow_pip_threshold_team0(self, shadow_pip_threshold_team0: float):
        await self.write_value_to_offset(480, shadow_pip_threshold_team0, "float")

    async def shadow_pip_threshold_team1(self) -> float:
        return await self.read_value_from_offset(484, "float")

    async def write_shadow_pip_threshold_team1(self, shadow_pip_threshold_team1: float):
        await self.write_value_to_offset(484, shadow_pip_threshold_team1, "float")

    async def scalar_damage(self) -> float:
        return await self.read_value_from_offset(512, "float")

    async def write_scalar_damage(self, scalar_damage: float):
        await self.write_value_to_offset(512, scalar_damage, "float")

    async def scalar_resist(self) -> float:
        return await self.read_value_from_offset(516, "float")

    async def write_scalar_resist(self, scalar_resist: float):
        await self.write_value_to_offset(516, scalar_resist, "float")

    async def scalar_pierce(self) -> float:
        return await self.read_value_from_offset(520, "float")

    async def write_scalar_pierce(self, scalar_pierce: float):
        await self.write_value_to_offset(520, scalar_pierce, "float")

    async def damage_limit(self) -> float:
        return await self.read_value_from_offset(524, "float")

    async def write_damage_limit(self, damage_limit: float):
        await self.write_value_to_offset(524, damage_limit, "float")

    async def d_k0(self) -> float:
        return await self.read_value_from_offset(528, "float")

    async def write_d_k0(self, d_k0: float):
        await self.write_value_to_offset(528, d_k0, "float")

    async def d_n0(self) -> float:
        return await self.read_value_from_offset(532, "float")

    async def write_d_n0(self, d_n0: float):
        await self.write_value_to_offset(532, d_n0, "float")

    async def resist_limit(self) -> float:
        return await self.read_value_from_offset(536, "float")

    async def write_resist_limit(self, resist_limit: float):
        await self.write_value_to_offset(536, resist_limit, "float")

    async def r_k0(self) -> float:
        return await self.read_value_from_offset(540, "float")

    async def write_r_k0(self, r_k0: float):
        await self.write_value_to_offset(540, r_k0, "float")

    async def r_n0(self) -> float:
        return await self.read_value_from_offset(544, "float")

    async def write_r_n0(self, r_n0: float):
        await self.write_value_to_offset(544, r_n0, "float")


class CurrentDuel(Duel):
    async def read_base_address(self) -> int:
        return await self.hook_handler.read_current_duel_base()
